import os
import re

import click

from rc3.commands import cmd_request
from rc3.common import json_helper, print_helper


@click.command("list", short_help="List collections, envs, requests.")
@click.argument('_type', type=str, required=False)
def cli(_type):
    """\b
    Defaults to listing all collections, and envs & requests in the current_collection

    \b
    TYPE is optional.
    And if supplied should be 1 of:
    1. collections, c
    2. environments, e
    3. requests, r
    """

    if _type is None or re.match('^all', _type, re.I):
        list_collections()
        list_environments()
        list_requests()
    elif re.match('^c', _type, re.I):
        list_collections()
    elif re.match('^e', _type, re.I):
        list_environments()
    elif re.match('^r', _type, re.I):
        list_requests()


def list_collections():
    _list = json_helper.read_collection_list()

    if len(_list) == 0:
        click.echo("No COLLECTIONS found in settings.json")
        raise click.Abort()
    click.echo("Listing COLLECTIONS found in settings.json:")

    # now display table
    header = ['NUM:', 'NAME:', 'LOCATION:']
    fields = ['display_num', 'name', 'location']
    print_helper.print_formatted_table(header, fields, _list)


def list_environments():
    _list = json_helper.read_environment_list()

    if len(_list) == 0:
        click.echo("No ENVIRONMENTS found in current_collection")
        raise click.Abort()
    click.echo("Listing ENVIRONMENTS found in current_collection:")

    # now display table
    header = ['NUM:', 'NAME:', 'baseUrl:']
    fields = ['display_num', 'name', 'baseUrl']
    print_helper.print_formatted_table(header, fields, _list)


def list_requests():
    _list = json_helper.read_request_list()

    if len(_list) == 0:
        click.echo("No REQUESTS found in current_collection")
        raise click.Abort()
    click.echo("Listing REQUESTS found in current_collection:")

    # now display table
    header = ['NUM:', 'FOLDER:', 'METHOD:', 'NAME:']
    fields = ['display_num', 'folder', 'method', 'name']
    print_helper.print_formatted_table(header, fields, _list)




import shutil
from importlib import resources
from importlib.resources import files

DATA_PACKAGE = 'rc3.data'
SCHEMA_BASE_URL = "http://localhost:8000"
SCHEMA_PREFIX = "rc3"
SCHEMA_VERSION = "0.0.3"
COLLECTION_FILENAME = "rc-collection.json"
FOLDER_FILENAME = "rc-folder.json"
SETTINGS_FILENAME = "rc-settings.json"
GLOBAL_ENV_FILENAME = "rc-global.json"


def get_file(filename):
    return files(DATA_PACKAGE).joinpath(filename)


def get_schema_file(partial):
    return get_file(f'home/schemas/{SCHEMA_PREFIX}-' + partial + f'-{SCHEMA_VERSION}.json')


def copy_tree(source_file, dest):
    source = get_file(source_file)
    with resources.as_file(source) as path:
        shutil.copytree(path, dest, dirs_exist_ok=True)


def copy(source_file, dest):
    source = get_file(source_file)
    with resources.as_file(source) as path:
        shutil.copy(path, dest)



import os

import pytest


# app code
def func(x):
    return x + 1


def f():
    raise SystemExit(1)


# tests
class TestClass:
    def test_answer(self):
        assert func(3) == 4

    def test_mytest(self):
        with pytest.raises(SystemExit):
            f()


def test_answer():
    assert func(4) == 5


def test_mytest(tmp_path):
    print("tmp_path.name")
    print(tmp_path.name)
    print(tmp_path)
    assert tmp_path.exists()
    d = tmp_path / "sub"
    d.mkdir()
    print(d)
    assert d.exists()



import os
from pathlib import Path

import pytest
from click.testing import CliRunner

from rc3 import cli


@pytest.fixture(scope="function")
def runner(request):
    return CliRunner()


@pytest.fixture(scope="function")
def clean_home(tmp_path, monkeypatch, autouse=True):
    home = tmp_path
    os.chdir(home)
    monkeypatch.setenv('RC_NO_CACHE', "True")
    monkeypatch.setenv('HOME', str(home))
    monkeypatch.setattr(Path, "home", lambda: Path(home))
    monkeypatch.setattr(os.path, 'expanduser', lambda u: home)
    yield home


@pytest.fixture(scope="function")
def clean_rc(clean_home, monkeypatch):
    rc_home = os.path.join(clean_home, '.rc')
    os.mkdir(rc_home)
    monkeypatch.setenv('RC_HOME', rc_home)
    yield rc_home


@pytest.fixture(scope="function", autouse=True)
def clean_empty(clean_home):
    empty_dir = os.path.join(clean_home, 'empty')
    os.mkdir(empty_dir)
    yield empty_dir


@pytest.fixture(scope="function")
def example_collection(clean_empty, runner):
    os.chdir(clean_empty)
    result = runner.invoke(cli.cli, ['init'])
    assert result.exit_code == 0
    yield clean_empty



import json
import os
import re
import click

from rc3.common import json_helper, print_helper


def process_subs(wrapper):
    if has_vars(wrapper):
        sub_vars(wrapper)
    # print_helper.print_json(r.get('_original', None))


def sub_vars(wrapper):
    envs = [
        json_helper.read_environment('current')[1],
        json_helper.read_environment('global')[1],
        os.environ
    ]
    r = wrapper.get('_original')

    # sub dicts
    sub_in_dict(envs, r.get('form_data'))
    sub_in_dict(envs, r.get('headers'))
    sub_in_dict(envs, r.get('params'))
    sub_in_dict(envs, r.get('auth'))

    # sub strings (& json body)
    r['url'] = sub_in_string(envs, r.get('url'))
    text = r.get('body', {}).get('text')
    if text is not None:
        r.get('body')['text'] = sub_in_string(envs, text)
    _json = r.get('body', {}).get('json')
    if _json is not None:
        json_string = json.dumps(_json)
        new_string = sub_in_string(envs, json_string)
        if new_string != json_string:
            r.get('body')['json'] = json.loads(new_string)


def lookup_var_value(envs, var):
    for env in envs:
        if var in env:
            return env.get(var)
    raise click.ClickException(f'var {{{{{var}}}}} is in the REQUEST but cannot be found in the current, global, or OS environment')


def sub_in_dict(envs, d):
    if d is None:
        return
    pattern = re.compile(r'{{(.*?)}}')
    for key, value in d.items():
        new_value = value
        for match in pattern.finditer(value):
            var = match.group(1).strip()
            var_value = lookup_var_value(envs, var)
            # this allows multiple vars to be used in a single value (each gets replaced)
            new_value = new_value.replace(match.group(0), var_value)
        d[key] = new_value


def sub_in_string(envs, s):
    if s is None:
        return None
    pattern = re.compile(r'{{(.*?)}}')
    for match in pattern.finditer(s):
        var = match.group(1).strip()
        var_value = lookup_var_value(envs, var)
        s = s.replace(match.group(0), var_value)
    return s


def has_vars(wrapper):
    r = wrapper.get('_original')
    dicts = [
        r.get('form_data',{}),
        r.get('headers',{}),
        r.get('params',{}),
        r.get('auth',{})
    ]
    strings = [
        r.get('url',''),
        r.get('body', {}).get('text',''),
        json.dumps(r.get('body', {}).get('json',{}))
    ]
    for d in dicts:
        for v in d.values():
            strings.append(v)

    pattern = re.compile(r'{{(.*?)}}')
    for s in strings:
        match = pattern.search(s)
        if match is not None:
            return True
    return False



import os
import re

import click
import pytest

from rc3 import cli
from rc3.commands import cmd_request
from rc3.common import json_helper, print_helper
from tests.commands import test_request
from tests.util.decorators import activate_responses, activate_recorder


def setup_localhost(runner):
    result = runner.invoke(cli.cli, ['e', '2'])
    assert result.exit_code == 0


def lookup_current():
    wrapper = cmd_request.lookup_request(None)
    r = wrapper.get('_original')
    return r, wrapper


def lookup_current_response():
    r, wrapper = lookup_current()
    response_dir = wrapper.get('_dir')
    response_file = wrapper.get('_filename').split('.')[0] + ".response"
    response_full_file = os.path.join(response_dir, response_file)
    if os.path.exists(response_full_file):
        response = json_helper.read_json(response_full_file)
        return response
    return None


# @activate_recorder()
@activate_responses()
def test_send_1(example_collection, runner):
    setup_localhost(runner)

    result = runner.invoke(cli.cli, ['send'])
    assert result.exit_code == 0
    assert "Hello" in result.output
    assert "English" in result.output

    response = lookup_current_response()
    assert response.get('status_code') == 200


# @activate_recorder()
@activate_responses()
def test_send_basics(example_collection, runner):
    setup_localhost(runner)

    result = runner.invoke(cli.cli, ['r', '--pick', '2'])
    assert result.exit_code == 0
    result = runner.invoke(cli.cli, ['send'])
    assert result.exit_code == 0
    response = lookup_current_response()
    assert response.get('status_code') == 200

    result = runner.invoke(cli.cli, ['r', '--pick', '3'])
    assert result.exit_code == 0
    result = runner.invoke(cli.cli, ['send'])
    assert result.exit_code == 0
    response = lookup_current_response()
    assert response.get('status_code') == 200

    result = runner.invoke(cli.cli, ['r', '--pick', '4'])
    assert result.exit_code == 0
    result = runner.invoke(cli.cli, ['send'])
    assert result.exit_code == 0
    response = lookup_current_response()
    assert response.get('status_code') == 200

    result = runner.invoke(cli.cli, ['r', '--pick', '5'])
    assert result.exit_code == 0
    result = runner.invoke(cli.cli, ['send'])
    assert result.exit_code == 0
    response = lookup_current_response()
    assert response.get('status_code') == 200


def test_var_missing(example_collection, runner):
    setup_localhost(runner)

    result = runner.invoke(cli.cli, ['send', '--pick', '6'])
    assert result.exit_code == 1
    assert "var {{token}} is in the REQUEST but cannot be found" in result.output


# @activate_recorder()
@activate_responses()
def test_mint_extract_use_token(example_collection, runner):
    setup_localhost(runner)

    # pre-test, confirm token doesn't exist/extract works
    result = runner.invoke(cli.cli, ['send', '--pick', '6'])
    assert result.exit_code == 1
    assert "var {{token}} is in the REQUEST but cannot be found" in result.output

    # mint and extract
    result = runner.invoke(cli.cli, ['send', '--pick', 'mint-admin-token'])
    assert result.exit_code == 0
    response = lookup_current_response()
    # NOTE: mint-admin-token, has save_responses=False, so lookup method will return None!
    assert response is None

    # use token
    result = runner.invoke(cli.cli, ['send', '--pick', '6'])
    assert result.exit_code == 0
    response = lookup_current_response()
    assert response.get('status_code') == 200


# @activate_recorder()
@activate_responses()
def test_not_verbose(example_collection, runner):
    setup_localhost(runner)

    result = runner.invoke(cli.cli, ['request', '--pick', '1'])
    assert result.exit_code == 0
    result = runner.invoke(cli.cli, ['send'])
    assert result.exit_code == 0

    response = lookup_current_response()
    assert print_helper.get_json_string(response).strip() not in result.output


# @activate_recorder()
@activate_responses()
def test_is_verbose(example_collection, runner):
    setup_localhost(runner)

    result = runner.invoke(cli.cli, ['request', '--pick', '1'])
    assert result.exit_code == 0
    result = runner.invoke(cli.cli, ['-v', 'send'])
    assert result.exit_code == 0

    response = lookup_current_response()
    assert print_helper.get_json_string(response).strip() in result.output


# @activate_recorder()
@activate_responses()
def test_request_no_responses(runner, example_collection, monkeypatch):
    setup_localhost(runner)

    # Setup global settings WITH responses
    settings = json_helper.read_settings()
    settings['save_responses'] = True
    json_helper.write_settings(settings)

    # Setup request, with no_responses (should override the global settings)
    r, wrapper = lookup_current()
    r['save_responses'] = False
    monkeypatch.setattr(click, "edit", lambda x: print_helper.get_json_string(r))
    # Setup continue, with save to file
    result = runner.invoke(cli.cli, ['r', "--edit"])
    assert result.exit_code == 0
    assert "REQUEST saved" in result.output

    result = runner.invoke(cli.cli, ['send'])
    assert result.exit_code == 0

    # .response file should NOT exist
    response = lookup_current_response()
    assert response is None


# @activate_recorder()
@activate_responses()
def test_settings_no_responses(runner, example_collection, monkeypatch):
    setup_localhost(runner)

    # Setup global settings, with no_responses
    settings = json_helper.read_settings()
    settings['save_responses'] = False
    json_helper.write_settings(settings)

    result = runner.invoke(cli.cli, ['send'])
    assert result.exit_code == 0

    # .response file should NOT exist
    response = lookup_current_response()
    assert response is None


# @activate_recorder()
@activate_responses()
def test_settings_with_responses(runner, example_collection, monkeypatch):
    setup_localhost(runner)

    # Setup global settings
    settings = json_helper.read_settings()
    settings['save_responses'] = True
    json_helper.write_settings(settings)

    result = runner.invoke(cli.cli, ['send'])
    assert result.exit_code == 0

    # .response file SHOULD exist
    response = lookup_current_response()
    assert response is not None
    assert response.get('status_code') == 200






import json
import os
import click

from rc3.commands import cmd_list
from rc3.common import json_helper, print_helper, config_helper
from rc3.common.data_helper import SCHEMA_BASE_URL, SCHEMA_PREFIX, SCHEMA_VERSION


@click.command("global", short_help="Manage the GLOBAL ENVIRONMENT stored at RC_HOME.")
@click.option('-i', '--info', is_flag=True, default=False, help="Display json of GLOBAL ENVIRONMENT.")
@click.option('-e', '--edit', is_flag=True, default=False, help="Edit the GLOBAL ENVIRONMENT with system editor.")
def cli(info, edit):
    """\b
    Manage the GLOBAL ENVIRONMENT stored at RC_HOME.

    """
    if edit:
        edit_environment()
    elif info:
        print_info()
    else:
        print_info()


def edit_environment():
    env_filename, env = json_helper.read_environment('global')
    new = internal_edit(env)
    if new is not None:
        json_helper.write_environment(env_filename, env)


def internal_edit(env):
    json_string = print_helper.get_json_string(env)
    new_string = click.edit(json_string)
    if new_string is not None:
        new_e = json_helper.parse_json(new_string)
        if new_e is None:
            raise click.ClickException("new ENVIRONMENT must be valid JSON.")
        if json_helper.validate(new_e, 'environment'):
            return new_e
    else:
        return None


def print_info():
    env_filename, env = json_helper.read_environment('global')
    print_helper.print_json(env)



import click

from rc3.common import json_helper


@click.command("root", short_help="Prints CURRENT_COLLECTION directory.")
def cli():
    """\b
    Will print the current collection directory (from settings.json) to STDOUT.

    \b
    I recommend editing collection & request json in VSCode.
    You should be able to launch VSCode @ collection root with:
    code $(rc root)
    """

    current_collection = json_helper.find_current_collection()
    if current_collection is None:
        print("No current_collection selected!")
    else:
        print(current_collection.get('location', "No location set in current_collection!"))



import click

from rc3.common import config_helper


@click.command("home", short_help="Prints RC_HOME directory.")
def cli():
    """\b
    Will print the RC_HOME directory to STDOUT.

    \b
    I recommend editing settings & global env json in VSCode.
    You should be able to launch VSCode @ RC_HOME root with:
    code $(rc home)
    """

    home = config_helper.get_config_folder()
    print(home)



import os
import re

import click
import pytest

from rc3 import cli
from rc3.commands import cmd_request, cmd_environment
from rc3.common import json_helper, print_helper


def test_new_qa(example_collection, runner):
    result = runner.invoke(cli.cli, ['e', '--new'], input="qa\n\n")
    assert result.exit_code == 0
    assert "ENVIRONMENT created" in result.output


def test_overwrite_qa(example_collection, runner):
    result = runner.invoke(cli.cli, ['e', '--new'], input="qa\n\n")
    assert result.exit_code == 0
    assert "ENVIRONMENT created" in result.output

    result = runner.invoke(cli.cli, ['e', '--new'], input="qa\ny\n\n")
    assert result.exit_code == 0
    assert "An ENVIRONMENT named 'qa' already exists" in result.output
    assert "ENVIRONMENT created" in result.output


def lookup_current():
    wrapper = cmd_environment.lookup_environment(None)
    e = wrapper.get('_original')
    return e, wrapper


def test_edit_nosave(example_collection, runner, monkeypatch):
    e, wrapper = lookup_current()
    monkeypatch.setattr(click, "edit", lambda x: None)

    # edit current
    result = runner.invoke(cli.cli, ['environment', "--edit"])
    assert result.exit_code == 0
    assert result.output == ''
    e2, wrapper2 = lookup_current()
    assert print_helper.get_json_string(e) == print_helper.get_json_string(e2)


def test_edit_withsave(runner, example_collection, monkeypatch):
    e, wrapper = lookup_current()
    e['baseUrl'] = "not even a url"
    monkeypatch.setattr(click, "edit", lambda x: print_helper.get_json_string(e))

    # edit with SAVE should save
    result = runner.invoke(cli.cli, ['environment', "--edit"])
    assert result.exit_code == 0
    assert "ENVIRONMENT saved" in result.output

    # new file should have the new value
    e2, wrapper2 = lookup_current()
    assert print_helper.get_json_string(e) == print_helper.get_json_string(e2)


def test_edit_errors(runner, example_collection, monkeypatch):
    e, wrapper = lookup_current()
    # invalid, schema calls for all strings as values
    e['bob'] = 2112
    monkeypatch.setattr(click, "edit", lambda x: print_helper.get_json_string(e))

    # edit with SAVE should VALIDATE JSON against schema and reject errors
    result = runner.invoke(cli.cli, ['e', "--edit"])
    assert result.exit_code == 0
    assert "JSON is invalid" in result.output
    assert "2112 is not of" in result.output

    # new file should NOT have the edits
    e2, wrapper2 = lookup_current()
    assert print_helper.get_json_string(e) != print_helper.get_json_string(e2)


def test_edit_unknown(runner, example_collection):
    result = runner.invoke(cli.cli, ['e', "--edit", "10"])
    assert result.exit_code == 1
    assert "not found" in result.output


def test_info(runner, example_collection):
    e, wrapper = lookup_current()

    result = runner.invoke(cli.cli, ['e', "--info"])
    assert result.exit_code == 0
    assert print_helper.get_json_string(e) in result.output


def test_no_options_list(runner, example_collection):
    result = runner.invoke(cli.cli, ['e'])
    assert result.exit_code == 0

    result2 = runner.invoke(cli.cli, ['e', '--list'])
    assert result2.exit_code == 0
    assert result.output == result2.output


def test_no_options_pick(runner, example_collection):
    e, wrapper = lookup_current()

    result = runner.invoke(cli.cli, ['e', '2'])
    assert result.exit_code == 0
    assert "ENVIRONMENT has been picked" in result.output

    e2, wrapper2 = lookup_current()
    assert print_helper.get_json_string(e) != print_helper.get_json_string(e2)

    result = runner.invoke(cli.cli, ['e', '--list'])
    assert result.exit_code == 0
    assert len(re.findall(r'1\*', result.output)) == 0
    assert len(re.findall(r'2\*', result.output)) == 1


def test_pick_prompt_bad_choice(runner, example_collection):
    result = runner.invoke(cli.cli, ['e', '--pick'], input="0\n")
    assert result.exit_code == 1
    assert "Invalid selection" in result.output


def test_pick_prompt(runner, example_collection):
    result = runner.invoke(cli.cli, ['e', '--pick'], input="2\n")
    assert result.exit_code == 0
    assert "ENVIRONMENT has been picked" in result.output

    result = runner.invoke(cli.cli, ['e', '--list'])
    assert result.exit_code == 0
    assert len(re.findall(r'1\*', result.output)) == 0
    assert len(re.findall(r'2\*', result.output)) == 1


def test_pick_number_works(runner, example_collection):
    result = runner.invoke(cli.cli, ['e', '--pick', '2'])
    assert result.exit_code == 0
    assert "ENVIRONMENT has been picked" in result.output

    result = runner.invoke(cli.cli, ['e', '--list'])
    assert result.exit_code == 0
    assert len(re.findall(r'1\*', result.output)) == 0
    assert len(re.findall(r'2\*', result.output)) == 1


def test_pick_name_works(runner, example_collection):
    result = runner.invoke(cli.cli, ['e', '--pick', 'localhost'])
    assert result.exit_code == 0
    assert "ENVIRONMENT has been picked" in result.output

    result = runner.invoke(cli.cli, ['e', '--list'])
    assert result.exit_code == 0
    assert len(re.findall(r'1\*', result.output)) == 0
    assert len(re.findall(r'2\*', result.output)) == 1


def test_pick_bad_values_error(runner, example_collection):
    result = runner.invoke(cli.cli, ['e', '--pick', '11'])
    assert result.exit_code == 1
    assert "not found" in result.output

    result = runner.invoke(cli.cli, ['e', '--pick', 'not-there'])
    assert result.exit_code == 1
    assert "not found" in result.output

    result = runner.invoke(cli.cli, ['e', '--list'])
    assert result.exit_code == 0
    assert len(re.findall(r'1\*', result.output)) == 1








import click

from rc3.common import json_helper, print_helper

@click.command("settings", short_help="Manage the GLOBAL RC SETTINGS stored at RC_HOME.")
@click.option('-i', '--info', is_flag=True, default=False, help="Display json of RC-SETTINGS.")
@click.option('-e', '--edit', is_flag=True, default=False, help="Edit the RC-SETTINGS file with system editor.")
def cli(info, edit):
    """\b
    Manage the GLOBAL RC-SETTINGS stored at RC_HOME.

    Alternatively, I recommend editing rc-settings & rc-global env in VSCode,
    where schema hints & validations will be shown.

    """

    if edit:
        edit_settings()
    elif info:
        print_info()
    else:
        print_info()


def edit_settings():
    settings = json_helper.read_settings()
    new = internal_edit(settings)
    if new is not None:
        json_helper.write_settings(new)


def internal_edit(settings):
    json_string = print_helper.get_json_string(settings)
    new_string = click.edit(json_string)
    if new_string is not None:
        new = json_helper.parse_json(new_string)
        if new is None:
            raise click.ClickException("new SETTINGS must be valid JSON.")
        if json_helper.validate(new, 'settings'):
            return new
    else:
        return None


def print_info():
    settings = json_helper.read_settings()
    print_helper.print_json(settings)



import os
import shutil

import click
from rc3.common import config_helper, json_helper, data_helper
from rc3.common.data_helper import SETTINGS_FILENAME, COLLECTION_FILENAME, GLOBAL_ENV_FILENAME


@click.command("init", short_help="Init RC settings, and Collection files.")
def cli():
    """\b
    Will do 0-4 of the following:
    1. Will create the RC_HOME directory if it doesn't exist.
    2. Will create RC_HOME/settings, global env, and schemas dir if they don't exist.
    3. Will initialize a new example Collection if ran from an empty directory.
    4. Will import the current directory if it contains a valid rc-collection.json file.
    """

    # do the 0-4 things
    init_rc_home()
    init_cwd_as_collection()
    import_collection()


def init_cwd_as_collection():
    cwd = os.getcwd()
    if not os.listdir(cwd):
        print("CWD is empty, creating sample Collection here: " + cwd)
        data_helper.copy_tree('collection', cwd)
    else:
        print("CWD is not empty, skipping collection creation...")


def import_collection():
    cwd = os.getcwd()
    print("Importing collection from: " + cwd)
    collection_dict = json_helper.load_and_validate(COLLECTION_FILENAME, _dir=cwd)
    if collection_dict is None:
        return

    # for any .defaults environments in the collection, copy them to .json (i.e. make a real environment)
    # note: there should also be a .gitignore in the env folder, so .json doesn't get committed
    env_folder = os.path.join(cwd, 'environments')
    defaults_counter = 0
    for dirpath, dirnames, files in os.walk(env_folder):
        for file in files:
            if file.endswith('.defaults'):
                name = file.split('.')[-2]
                default_file = os.path.join(dirpath, file)
                new_file = os.path.join(dirpath, name + ".json")
                if not os.path.exists(new_file):
                    defaults_counter += 1
                    shutil.copy(default_file, new_file)
    if defaults_counter > 0:
        print(f"{defaults_counter} default environment(s) initialized in your collection")

    # get "name" from json, or use cwd directory name
    parts = os.path.split(cwd)
    name = collection_dict.get("name", parts[-1])

    settings = json_helper.read_settings()
    settings["current_collection"] = name
    # ONLY add to the collections list if name is NOT already there!
    if any(filter(lambda c: c['name'] == name, settings['collections'])):
        print("Name already exists in global settings: " + name)
    else:
        print("Adding collection to global settings: " + name)
        settings["collections"].append({
            "name": name,
            "location": cwd
        })
    json_helper.write_settings(settings)


def init_rc_home():
    # Note: this next cmd will always create RC_HOME / ~/.rc if it doesn't exist
    home = config_helper.get_config_folder()

    dest = os.path.join(home, SETTINGS_FILENAME)
    if not os.path.exists(dest):
        print("Creating " + dest)
        data_helper.copy('home/' + SETTINGS_FILENAME, dest)

    dest = os.path.join(home, GLOBAL_ENV_FILENAME)
    if not os.path.exists(dest):
        print("Creating " + dest)
        data_helper.copy('home/' + GLOBAL_ENV_FILENAME, dest)

    dest = os.path.join(home, 'schemas')
    if not os.path.exists(dest):
        print("Creating " + dest)
        data_helper.copy_tree('home/schemas', dest)



import json
import os
import click

from rc3.commands import cmd_list
from rc3.common import json_helper, print_helper
from rc3.common.data_helper import COLLECTION_FILENAME


@click.command("collection", short_help="Manage COLLECTIONS defined in settings.")
@click.option('-l', '--list', '_list', is_flag=True, default=False, help="List COLLECTIONS in settings.")
@click.option('-i', '--info', is_flag=True, default=False, help="Display json of current COLLECTION.")
@click.option('-p', '--pick', is_flag=True, default=False, help="Pick an item as current COLLECTION.")
@click.option('-e', '--edit', is_flag=True, default=False, help="Edit a COLLECTION with system editor.")
@click.argument('collection_name', type=str, required=False)
def cli(_list, info, pick, edit, collection_name):
    """\b
    Manages COLLECTIONS that have been imported into settings.json.

    \b
    COLLECTION_NAME is optional.
    COLLECTION_NAME will default to the current_collection, or prompt you to pick.
    COLLECTION_NAME if used should be one of:
    1. The NUM column from 'rc collection --list' output
    2. THe NAME column from 'rc collection --list' output

    """
    if _list:
        cmd_list.list_collections()
    elif edit:
        click.echo("NOT IMPLEMENTED! Please use VSCode to edit your collection.")
    elif info:
        print_info(collection_name)
    elif pick:
        pick_collection(collection_name)
    elif collection_name is None:
        # DEFAULT to list() if no COLLECTION_NAME
        cmd_list.list_collections()
    else:
        # DEFAULT to pick() if COLLECTION_NAME
        pick_collection(collection_name)


def pick_collection(name=None):
    c = lookup_or_prompt(name)
    if c is None:
        return None

    settings = json_helper.read_settings()
    settings['current_collection'] = c['name']
    json_helper.write_settings(settings)

    click.echo("COLLECTION has been picked: " + c['name'])
    return c


def lookup_or_prompt(name, prompt="Which COLLECTION do you want to pick?"):
    _list = json_helper.read_collection_list()
    if len(_list) == 0:
        raise click.ClickException("Invalid option.  There are no COLLECTIONS defined.")

    picked_object = None
    if name is None:
        cmd_list.list_collections()
        current_c = lookup_collection(None)
        default_choice = 0 if current_c is None else current_c.get('number', 0)
        number = click.prompt(prompt, default=default_choice)
        picked_object = lookup_collection(str(number))
        if picked_object is None:
            raise click.ClickException("Invalid selection")
    else:
        picked_object = lookup_collection(name)
        if picked_object is None:
            raise click.ClickException("COLLECTION_NAME not found: " + name)
        else:
            click.echo("Found COLLECTION_NAME: " + name)
    return picked_object


def print_info(name):
    wrapper = lookup_collection(name)
    c = json_helper.load_and_validate(COLLECTION_FILENAME, _dir=wrapper.get('location'))
    print_helper.print_json(c)


def lookup_collection(name):
    _list = json_helper.read_collection_list()

    for c in _list:
        if name is None and c.get('_current', False):
            return c
        if name is not None and name == c.get('name', None):
            return c
        if name is not None and name == str(c.get('number', 0)):
            return c
    else:
        return None



import os
import re

import click
import requests
from jsonpath_ng import parse
from requests.auth import HTTPBasicAuth, AuthBase

from rc3.commands import cmd_request
from rc3.common import json_helper, print_helper, env_helper, inherit_helper


@click.command("subs", short_help="Temp test subs.")
@click.option('-p', '--pick', is_flag=True, default=False, help="Pick a REQUEST then sub it.")
@click.argument('request_name', type=str, required=False)
def cli(pick, request_name):
    """\b
    Will sub an REQUEST object.
    """
    if pick:
        r = cmd_request.pick_request()
        inherit_helper.find_auth(r)
        # env_helper.process_subs(r)
    else:
        r = lookup_request(request_name)
        inherit_helper.find_auth(r)
        # env_helper.process_subs(r)


def lookup_request(request_name):
    r = cmd_request.lookup_request(request_name)
    if r is None and request_name is None:
        raise click.ClickException("There is no current REQUEST, exiting...")
    if r is None:
        raise click.ClickException("REQUEST '{}' not found. See 'rc request --list'".format(request_name))
    return r



import os
import re

import click
import pytest

from rc3 import cli
from rc3.commands import cmd_request
from rc3.common import json_helper, print_helper


def test_not_edit(example_collection, runner):
    result = runner.invoke(cli.cli, ['c', '--edit'])
    assert result.exit_code == 0
    assert "NOT IMPLEMENTED!" in result.output


def test_info(example_collection, runner):
    result = runner.invoke(cli.cli, ['c', '--info'])
    assert result.exit_code == 0
    assert "name" in result.output
    assert "current_request" in result.output
    assert "current_environment" in result.output



import json
import os
import re
import click

from rc3.common import json_helper, print_helper
from rc3.common.data_helper import FOLDER_FILENAME, COLLECTION_FILENAME


def find_auth(wrapper):
    r = wrapper.get('_original')
    if r.get('auth', {}).get('type','inherit') != 'inherit':
        return r.get('auth')
    auth = walk_up_folders(wrapper.get('_dir'))
    # print_helper.print_json(auth)
    return auth


def walk_up_folders(_dir):
    collection_filename = os.path.join(_dir, COLLECTION_FILENAME)
    if os.path.exists(collection_filename):
        # we;re at the root of the collection stop walking & return something!
        c, wrapper = json_helper.read_current_collection()
        if c.get('auth',{}).get('type','inherit') != 'inherit':
            return c.get('auth')
        else:
            return {
                "type": "none"
            }

    folder_filename = os.path.join(_dir, FOLDER_FILENAME)
    if os.path.exists(folder_filename):
        folder = json_helper.load_and_validate(FOLDER_FILENAME, _dir=_dir)
        if folder.get('auth',{}).get('type','inherit') == 'inherit':
            # RECURSIVE walk_up_folders!
            return walk_up_folders(os.path.join(_dir, os.pardir))
        else:
            return folder.get('auth')

    # RECURSIVE walk_up_folders!
    return walk_up_folders(os.path.join(_dir, os.pardir))



import os
import re

DEFAULT_CONFIG_FOLDER = '.rc'
RC_HOME = 'RC_HOME'


def get_config_folder():
    rc_home = os.getenv(RC_HOME)
    if rc_home is None:
        home = os.path.expanduser("~")
        rc_home = os.path.join(home, DEFAULT_CONFIG_FOLDER)
    if not os.path.exists(rc_home):
        os.mkdir(rc_home)
    return rc_home


def clean_filename(name):
    if name is None:
        return None
    name = name.lower()
    name = re.sub(' ', '_', name)
    name = re.sub('[^a-zA-Z0-9_-]', '', name)
    return name



import json
import os
import click

from rc3.commands import cmd_list
from rc3.common import json_helper, print_helper, config_helper
from rc3.common.data_helper import SCHEMA_BASE_URL, SCHEMA_PREFIX, SCHEMA_VERSION


@click.command("collection", short_help="Manage ENVIRONMENTS defined in a collection.")
@click.option('-l', '--list', '_list', is_flag=True, default=False, help="List ENVIRONMENTS in a collection.")
@click.option('-i', '--info', is_flag=True, default=False, help="Display json of current ENVIRONMENT.")
@click.option('-n', '--new', is_flag=True, default=False, help="Create a new ENVIRONMENT in a collection.")
@click.option('-p', '--pick', is_flag=True, default=False, help="Pick an item as current ENVIRONMENT.")
@click.option('-e', '--edit', is_flag=True, default=False, help="Edit an ENVIRONMENT with system editor.")
@click.argument('environment_name', type=str, required=False)
def cli(_list, info, new, pick, edit, environment_name):
    """\b
    Manages ENVIRONMENTS in the current_collection.

    \b
    ENVIRONMENT_NAME is optional.
    ENVIRONMENT_NAME will default to the current_environment, or prompt you to pick.
    ENVIRONMENT_NAME if used should be one of:
    1. The NUM column from 'rc environment --list' output
    2. THe NAME column from 'rc environment --list' output
    """
    if _list:
        cmd_list.list_environments()
    elif new:
        create_environment(environment_name)
    elif edit:
        edit_environment(environment_name)
    elif info:
        print_info(environment_name)
    elif pick:
        pick_environment(environment_name)
    elif environment_name is None:
        # DEFAULT to list() if no ENVIRONMENT_NAME
        cmd_list.list_environments()
    else:
        # DEFAULT to pick() if ENVIRONMENT_NAME
        pick_environment(environment_name)


def pick_environment(name=None):
    e = lookup_or_prompt(name)
    if e is None:
        return None

    c, c_wrapper = json_helper.read_current_collection()
    c['current_environment'] = e['name']
    json_helper.write_collection(c_wrapper)

    click.echo("ENVIRONMENT has been picked: " + e['name'])
    return e


def edit_environment(name, e=None):
    if e is None:
        e = lookup_environment(name)
    if e is None:
        raise click.ClickException("ENVIRONMENT '{}' not found. See 'rc environment --list'".format(name))
    new = internal_edit(e)
    if new is not None:
        e['_original'] = new
        save_environment(e, new)
    return e


def internal_edit(r):
    json_string = print_helper.get_json_string(r.get('_original'))
    new_string = click.edit(json_string)
    if new_string is not None:
        new_e = json_helper.parse_json(new_string)
        if new_e is None:
            raise click.ClickException("new ENVIRONMENT must be valid JSON.")
        if json_helper.validate(new_e, 'environment'):
            return new_e
    else:
        return None


def lookup_or_prompt(name, prompt="Which ENVIRONMENT do you want to pick?"):
    _list = json_helper.read_environment_list()
    if len(_list) == 0:
        raise click.ClickException("Invalid option.  There are no ENVIRONMENTS defined.")

    picked_object = None
    if name is None:
        cmd_list.list_environments()
        e = lookup_environment(None)
        default_choice = 0 if e is None else e.get('number', 0)
        number = click.prompt(prompt, default=default_choice)
        picked_object = lookup_environment(str(number))
        if picked_object is None:
            raise click.ClickException("Invalid selection")
    else:
        picked_object = lookup_environment(name)
        if picked_object is None:
            raise click.ClickException("ENVIRONMENT_NAME not found: " + name)
        else:
            click.echo("Found ENVIRONMENT_NAME: " + name)
    return picked_object


def print_info(name):
    e = lookup_environment(name)
    print_helper.print_json(e['_original'])


def lookup_environment(name):
    _list = json_helper.read_environment_list()

    for e in _list:
        if name is None and e.get('_current', False):
            return e
        if name is not None and name == e.get('name', None):
            return e
        if name is not None and name == str(e.get('number', 0)):
            return e
    else:
        return None


def save_environment(e, new):
    json_helper.write_environment(e['_filename'], new)
    print("ENVIRONMENT saved: " + e['name'])


def create_environment(name):
    name = config_helper.clean_filename(name)
    display_name = '' if name is None else name
    click.echo(f'Creating new ENVIRONMENT {display_name}')

    env = {
        "$schema": f'{SCHEMA_BASE_URL}/{SCHEMA_PREFIX}-environment-{SCHEMA_VERSION}.json',
        "baseUrl": "http://localhost:8080",
        "property1": "value1"
    }

    if name is None:
        name = click.prompt("Enter a short NAME for this ENVIRONMENT", default=display_name)
        name = config_helper.clean_filename(name)

    # Check if existing ENV, and confirm overwrite
    _list = json_helper.read_environment_list()
    for e in _list:
        if name == e['name']:
            if not click.confirm("An ENVIRONMENT named '{}' already exists, do you want to overwrite it?".format(name)):
                raise click.Abort()

    env['baseUrl'] = click.prompt("Enter the BASE URL for this ENVIRONMENT", default=env['baseUrl'])
    json_helper.write_environment(name+".json", env)
    print("ENVIRONMENT created: " + name)




"""
Rest Helper2 is a CLI tool used to make REST calls to APIS
It supports defining apps, envs, and resources which can be exported
as JSON, and imported by another user/machine, so they can easily call
the same REST apis.
"""
__version__ = "0.2"


import os
import re

import pytest

from rc3 import cli
from rc3.common import json_helper


def test_initial_list(example_collection, runner):
    result = runner.invoke(cli.cli, ['list'])
    assert result.exit_code == 0

    assert "Listing COLLECTIONS" in result.output
    assert "Listing ENVIRONMENTS" in result.output
    assert "Listing REQUESTS" in result.output
    assert len(re.findall(r'1\*', result.output)) == 3



import os

import pytest

from rc3 import cli


def test_first_run_help(runner):
    result = runner.invoke(cli.cli)
    assert result.exit_code == 0
    assert "Usage" in result.output
    assert "Show this message and exit" in result.output


def test_any_command_creates_empty_rc_home(clean_home, runner):
    # pre-test it doesn't exist from a pytest fixture
    rc_home = os.path.join(clean_home, '.rc')
    assert not os.path.exists(rc_home)

    # hello cmd doesn't matter, we just need some cmd some cli.cli is invoked instead of help displayed
    result = runner.invoke(cli.cli, ['hello'])
    assert result.exit_code == 0
    assert "Hello" in result.output

    # test it exists now AND is empty
    assert os.path.exists(rc_home)
    assert len(os.listdir(rc_home)) == 0


def test_shortcuts_work(example_collection, runner):
    result = runner.invoke(cli.cli, ['r', '--list'])
    assert result.exit_code == 0
    assert "Listing REQUESTS" in result.output
    result2 = runner.invoke(cli.cli, ['request', '--list'])
    assert result2.exit_code == 0
    assert result.output == result2.output

    result = runner.invoke(cli.cli, ['c', '--list'])
    assert result.exit_code == 0
    assert "Listing COLLECTIONS" in result.output
    result2 = runner.invoke(cli.cli, ['collection', '--list'])
    assert result2.exit_code == 0
    assert result.output == result2.output

    result = runner.invoke(cli.cli, ['e', '--list'])
    assert result.exit_code == 0
    assert "Listing ENVIRONMENTS" in result.output
    result2 = runner.invoke(cli.cli, ['environment', '--list'])
    assert result2.exit_code == 0
    assert result.output == result2.output



import click


@click.command("hello", short_help="Says hello so many times.")
@click.option('-c', '--count', default=1, help='Number of greetings')
@click.option('--title', help="Title to be used")
@click.option('--mr', is_flag=True, default=False, help="Shortcut for --title Mr.")
@click.argument('name', default="World")
def cli(count, title, mr, name):
    """Says hello."""
    if mr:
        title = "Mr."
    if title:
        name = title + " " + name
    for x in range(count):
        click.echo(f"Hello {name}!")



import os
import re

import pytest
from click import ClickException

from rc3 import cli
from rc3.common import json_helper, print_helper, env_helper
from tests.commands import test_request


def add_to_current(var, value):
    env_filename, env = json_helper.read_environment('current')
    env[var] = value
    json_helper.write_environment(env_filename, env)


def add_to_global(var, value):
    env_filename, env = json_helper.read_environment('global')
    env[var] = value
    json_helper.write_environment(env_filename, env)


def add_to_shell(var, value):
    os.environ[var] = value


def test_subbing_all_spots(example_collection, runner):
    r, wrapper = test_request.lookup_current()
    r['form_data'] = {
        "something": "{{ bob }}"
    }
    r['headers']["another"] = "{{ bob }}"
    r['params']["more"] = "{{ bob }}"
    r['auth']["username"] = "{{ bob }}"
    r['url'] = "{{ bob }}"
    r['body'] = {
        'text': "{{ bob }}",
        'json': {
            "some": "{{ bob }}",
            "more": "{{ bob }}"
        }
    }
    json_string = print_helper.get_json_string(r)
    assert len(re.findall(r'{{ bob }}', json_string)) == 8
    assert len(re.findall(r'current_bob', json_string)) == 0
    add_to_current("bob", "current_bob")

    env_helper.process_subs(wrapper)
    json_string = print_helper.get_json_string(r)
    assert len(re.findall(r'{{ bob }}', json_string)) == 0
    assert len(re.findall(r'current_bob', json_string)) == 8


def test_sub_current(example_collection, runner):
    r, wrapper = test_request.lookup_current()
    r['auth']["username"] = "{{ bob }}"

    json_string = print_helper.get_json_string(r)
    assert len(re.findall(r'{{ bob }}', json_string)) == 1
    assert len(re.findall(r'current_bob', json_string)) == 0
    add_to_current("bob", "current_bob")

    env_helper.process_subs(wrapper)

    json_string = print_helper.get_json_string(r)
    assert len(re.findall(r'{{ bob }}', json_string)) == 0
    assert len(re.findall(r'current_bob', json_string)) == 1


def test_sub_global(example_collection, runner):
    r, wrapper = test_request.lookup_current()
    r['auth']["username"] = "{{ bob }}"

    json_string = print_helper.get_json_string(r)
    assert len(re.findall(r'{{ bob }}', json_string)) == 1
    assert len(re.findall(r'global_bob', json_string)) == 0
    add_to_global("bob", "global_bob")

    env_helper.process_subs(wrapper)

    json_string = print_helper.get_json_string(r)
    assert len(re.findall(r'{{ bob }}', json_string)) == 0
    assert len(re.findall(r'global_bob', json_string)) == 1


def test_sub_shell(example_collection, runner):
    r, wrapper = test_request.lookup_current()
    r['auth']["username"] = "{{ bob }}"

    json_string = print_helper.get_json_string(r)
    assert len(re.findall(r'{{ bob }}', json_string)) == 1
    assert len(re.findall(r'galactic_bob', json_string)) == 0
    add_to_shell("bob", "galactic_bob")

    env_helper.process_subs(wrapper)

    json_string = print_helper.get_json_string(r)
    assert len(re.findall(r'{{ bob }}', json_string)) == 0
    assert len(re.findall(r'galactic_bob', json_string)) == 1


def test_missing_var(example_collection, runner):
    r, wrapper = test_request.lookup_current()
    r['auth']["username"] = "{{ missing_bob }}"

    with pytest.raises(ClickException, match=r'cannot be found in the current, global, or OS environment'):
        env_helper.process_subs(wrapper)


def test_current_wins(example_collection, runner):
    r, wrapper = test_request.lookup_current()
    r['auth']["username"] = "{{ bob }}"
    add_to_current("bob", "current_bob")
    add_to_global("bob", "global_bob")
    add_to_shell("bob", "galactic_bob")

    json_string = print_helper.get_json_string(r)
    assert len(re.findall(r'{{ bob }}', json_string)) == 1
    assert len(re.findall(r'current_bob', json_string)) == 0
    assert len(re.findall(r'global_bob', json_string)) == 0
    assert len(re.findall(r'galactic_bob', json_string)) == 0

    env_helper.process_subs(wrapper)

    json_string = print_helper.get_json_string(r)
    assert len(re.findall(r'{{ bob }}', json_string)) == 0
    assert len(re.findall(r'current_bob', json_string)) == 1
    assert len(re.findall(r'global_bob', json_string)) == 0
    assert len(re.findall(r'galactic_bob', json_string)) == 0


def test_global_wins(example_collection, runner):
    r, wrapper = test_request.lookup_current()
    r['auth']["username"] = "{{ bob }}"
    # add_to_current("bob", "current_bob")
    add_to_global("bob", "global_bob")
    add_to_shell("bob", "galactic_bob")

    json_string = print_helper.get_json_string(r)
    assert len(re.findall(r'{{ bob }}', json_string)) == 1
    assert len(re.findall(r'current_bob', json_string)) == 0
    assert len(re.findall(r'global_bob', json_string)) == 0
    assert len(re.findall(r'galactic_bob', json_string)) == 0

    env_helper.process_subs(wrapper)

    json_string = print_helper.get_json_string(r)
    assert len(re.findall(r'{{ bob }}', json_string)) == 0
    assert len(re.findall(r'current_bob', json_string)) == 0
    assert len(re.findall(r'global_bob', json_string)) == 1
    assert len(re.findall(r'galactic_bob', json_string)) == 0



import os
import re

import pytest

from rc3 import cli
from rc3.common import json_helper, config_helper
from rc3.common.data_helper import SETTINGS_FILENAME, COLLECTION_FILENAME, GLOBAL_ENV_FILENAME


def test_init_from_empty(clean_home, clean_empty, runner):
    # pre-test RC_HOME doesn't exist
    rc_home = os.path.join(clean_home, '.rc')
    assert not os.path.exists(rc_home)

    # change to empty dir, and run init
    os.chdir(clean_empty)
    result = runner.invoke(cli.cli, ['init'])
    assert result.exit_code == 0

    # test it exists now AND is empty
    assert os.path.exists(rc_home)
    assert os.listdir(rc_home) == [GLOBAL_ENV_FILENAME, SETTINGS_FILENAME, 'schemas']
    assert os.listdir(clean_empty) == ['environments',
                                       'examples',
                                       'greetings-basic',
                                       'greetings-oauth2',
                                       COLLECTION_FILENAME]
    settings = json_helper.read_settings()
    assert settings.get('current_collection') == "example-collection"


def test_init_from_NOT_empty(clean_home, clean_empty, runner):
    # pre-test RC_HOME doesn't exist
    rc_home = os.path.join(clean_home, '.rc')
    assert not os.path.exists(rc_home)

    # change to empty dir, create tempfile
    # and then run init
    os.chdir(clean_empty)
    json_helper.write_json("temp.json",{})
    result = runner.invoke(cli.cli, ['init'])
    assert result.exit_code == 0

    # test that we DO STILL init RC_HOME
    assert os.path.exists(rc_home)
    assert os.listdir(rc_home) == [GLOBAL_ENV_FILENAME, SETTINGS_FILENAME, 'schemas']
    # BUT we DON'T init the CWD, or import a collection
    assert os.listdir(clean_empty) == ['temp.json']
    settings = json_helper.read_settings()
    assert settings.get('current_collection') == ""




import os
import re

import click
import pytest

from rc3 import cli
from rc3.commands import cmd_request
from rc3.common import json_helper, print_helper


def test_not_new(example_collection, runner):
    result = runner.invoke(cli.cli, ['request', '--new'])
    assert result.exit_code == 0
    assert "NOT IMPLEMENTED!" in result.output


def test_list_variations(example_collection, runner):
    result = runner.invoke(cli.cli, ['request', '--list'])
    assert result.exit_code == 0
    result2 = runner.invoke(cli.cli, ['request'])
    assert result2.exit_code == 0
    result3 = runner.invoke(cli.cli, ['r', '--list'])
    assert result3.exit_code == 0
    result4 = runner.invoke(cli.cli, ['r'])
    assert result4.exit_code == 0

    assert result.output == result2.output == result3.output == result4.output
    assert "Listing REQUESTS" in result.output
    assert len(re.findall(r'1\*', result.output)) == 1


def lookup_current():
    wrapper = cmd_request.lookup_request(None)
    r = wrapper.get('_original')
    return r, wrapper


def test_edit_nosave(example_collection, runner, monkeypatch):
    r, wrapper = lookup_current()
    monkeypatch.setattr(click, "edit", lambda x: None)

    # edit current REQUEST
    result = runner.invoke(cli.cli, ['r', "--edit"])
    assert result.exit_code == 0
    assert result.output == ''
    r2, wrapper2 = lookup_current()
    assert print_helper.get_json_string(r) == print_helper.get_json_string(r2)


def test_edit_withsave(runner, example_collection, monkeypatch):
    r, wrapper = lookup_current()
    r['method'] = "POST"
    monkeypatch.setattr(click, "edit", lambda x: print_helper.get_json_string(r))

    # edit with SAVE should save
    result = runner.invoke(cli.cli, ['r', "--edit"])
    assert result.exit_code == 0
    assert "REQUEST saved" in result.output

    # new file should have the new value
    r2, wrapper2 = lookup_current()
    assert print_helper.get_json_string(r) == print_helper.get_json_string(r2)


def test_edit_errors(runner, example_collection, monkeypatch):
    r, wrapper = lookup_current()
    r['method'] = "WHAT?"
    monkeypatch.setattr(click, "edit", lambda x: print_helper.get_json_string(r))

    # edit with SAVE should VALIDATE JSON against schema and reject errors
    result = runner.invoke(cli.cli, ['r', "--edit"])
    assert result.exit_code == 0
    assert "JSON is invalid" in result.output
    assert "'WHAT?' is not one of" in result.output

    # new file should NOT have the edits
    r2, wrapper2 = lookup_current()
    assert print_helper.get_json_string(r) != print_helper.get_json_string(r2)


def test_edit_unknown(runner, example_collection, monkeypatch):
    result = runner.invoke(cli.cli, ['r', "--edit", "10"])
    assert result.exit_code == 1
    assert "not found" in result.output


def test_info(runner, example_collection, monkeypatch):
    r, wrapper = lookup_current()

    result = runner.invoke(cli.cli, ['r', "--info"])
    assert result.exit_code == 0
    assert print_helper.get_json_string(r) in result.output


def test_no_options_list(runner, example_collection, monkeypatch):
    r, wrapper = lookup_current()

    result = runner.invoke(cli.cli, ['r'])
    assert result.exit_code == 0

    result2 = runner.invoke(cli.cli, ['r', '--list'])
    assert result2.exit_code == 0
    assert result.output == result2.output


def test_no_options_pick(runner, example_collection, monkeypatch):
    r, wrapper = lookup_current()

    result = runner.invoke(cli.cli, ['r', '3'])
    assert result.exit_code == 0
    assert "REQUEST has been picked" in result.output

    r2, wrapper2 = lookup_current()
    assert print_helper.get_json_string(r) != print_helper.get_json_string(r2)

    result = runner.invoke(cli.cli, ['r', '--list'])
    assert result.exit_code == 0
    assert len(re.findall(r'1\*', result.output)) == 0
    assert len(re.findall(r'3\*', result.output)) == 1


def test_pick_prompt_bad_choice(runner, example_collection, monkeypatch):
    _input = "0\n"
    result = runner.invoke(cli.cli, ['r', '--pick'], input=_input)
    assert result.exit_code == 1
    assert "Invalid selection" in result.output


def test_pick_prompt(runner, example_collection, monkeypatch):
    _input = "3\n"
    result = runner.invoke(cli.cli, ['r', '--pick'], input=_input)
    assert result.exit_code == 0
    assert "REQUEST has been picked" in result.output

    result = runner.invoke(cli.cli, ['r', '--list'])
    assert result.exit_code == 0
    assert len(re.findall(r'1\*', result.output)) == 0
    assert len(re.findall(r'3\*', result.output)) == 1


def test_pick_number_works(runner, example_collection, monkeypatch):
    result = runner.invoke(cli.cli, ['r', '--pick', '4'])
    assert result.exit_code == 0
    assert "REQUEST has been picked" in result.output

    result = runner.invoke(cli.cli, ['r', '--list'])
    assert result.exit_code == 0
    assert len(re.findall(r'1\*', result.output)) == 0
    assert len(re.findall(r'4\*', result.output)) == 1


def test_pick_name_works(runner, example_collection, monkeypatch):
    result = runner.invoke(cli.cli, ['r', '--pick', 'mint-admin-token'])
    assert result.exit_code == 0
    assert "REQUEST has been picked" in result.output

    result = runner.invoke(cli.cli, ['r', '--list'])
    assert result.exit_code == 0
    assert len(re.findall(r'1\*', result.output)) == 0
    assert len(re.findall(r'7\*', result.output)) == 1


def test_pick_bad_values_error(runner, example_collection, monkeypatch):
    result = runner.invoke(cli.cli, ['r', '--pick', '11'])
    assert result.exit_code == 1
    assert "not found" in result.output

    result = runner.invoke(cli.cli, ['r', '--pick', 'not-there'])
    assert result.exit_code == 1
    assert "not found" in result.output

    result = runner.invoke(cli.cli, ['r', '--list'])
    assert result.exit_code == 0
    assert len(re.findall(r'1\*', result.output)) == 1





import json
import os
import sys
from json import JSONDecodeError

import click
from funcy import print_durations
from jsonschema import Draft7Validator
from referencing import Registry, Resource
from rc3.common import config_helper, data_helper
from rc3.common.data_helper import SCHEMA_PREFIX, SCHEMA_VERSION, COLLECTION_FILENAME, SETTINGS_FILENAME, \
    FOLDER_FILENAME, GLOBAL_ENV_FILENAME
from rc3.common.decorators import rc_print_durations, rc_memoized


def guess_schema(filename):
    if filename == COLLECTION_FILENAME:
        return 'collection'
    if filename == SETTINGS_FILENAME:
        return 'settings'
    if filename == GLOBAL_ENV_FILENAME:
        return 'environment'
    if filename == FOLDER_FILENAME:
        return 'folder'
    if filename.endswith('.request'):
        return 'request'
    # default, because it's the only one that can't be guessed based on filename
    return 'environment'


@rc_print_durations
@rc_memoized
def guess_dir(filename):
    home = config_helper.get_config_folder()
    if filename in [SETTINGS_FILENAME, GLOBAL_ENV_FILENAME]:
        return home
    if filename in os.listdir(home):
        return home

    current_collection = find_current_collection()
    current_location = current_collection.get('location', None) or None
    if current_location is None:
        return None
    for dirpath, dirnames, files in os.walk(current_location):
        if filename in files:
            return dirpath

    if filename.endswith(".json"):
        return os.path.join(current_location, "environments")

    return None


@rc_print_durations
@rc_memoized
# @rc_memoized
def create_registry():
    # Registry is only needed for relative refs ($ref) in schemas (and we only have 1 to begin with)
    # All the full document schemas will be loaded individually when used
    registry = Registry()
    loaded = read_json(data_helper.get_schema_file('auth'))
    resource = Resource.from_contents(loaded)
    registry = registry.with_resource(f"./{SCHEMA_PREFIX}-auth-{SCHEMA_VERSION}.json", resource)
    return registry


def load_and_validate(filename, schema=None, _dir=None):
    # print("LOAD AND VALIDATE: " + filename)
    # if schema=None, guess based on filename
    if schema is None:
        schema = guess_schema(filename)
        # print(f'filename=[{filename}, schema=[{schema}]')

    # if _dir=None, look for filename in RC_HOME, or current_collection
    if _dir is None:
        _dir = guess_dir(filename)

    file = os.path.join(_dir, filename)
    if not os.path.exists(file):
        print("Can't find: " + filename)
        sys.exit()

    _dict = read_json(file)
    schema_dict = read_json(data_helper.get_schema_file(schema))

    # This PDF seemed better than the web docs for jsonschema:
    # https://readthedocs.org/projects/python-jsonschema/downloads/pdf/latest/
    validator = Draft7Validator(schema_dict, registry=create_registry())

    validator.check_schema(schema_dict)
    if validator.is_valid(_dict):
        return _dict

    print()
    print("Error: file doesn't pass JSON schema validation: " + file)
    # print(filename + " is invalid:")
    for error in validator.iter_errors(_dict):
        # print(vars(error)) # use to look at what other attributes are available...
        print(" * json path: /" + '.'.join(error.path))
        print(" * error: " + error.message)
    print("If needed, please use VSCode to see validation errors w/ squiggles & hovers")
    sys.exit()


def validate(_dict, schema=None):
    if schema is None:
        schema = 'request'

    schema_dict = read_json(data_helper.get_schema_file(schema))
    validator = Draft7Validator(schema_dict, registry=create_registry())

    validator.check_schema(schema_dict)
    if validator.is_valid(_dict):
        return _dict

    print("JSON is invalid: " + schema)
    for error in validator.iter_errors(_dict):
        # print(vars(error)) # use to look at what other attributes are available...
        print(" * json path: /" + '.'.join(error.path))
        print(" * error: " + error.message)
    print("If needed, please use VSCode to see validation errors w/ squiggles & hovers")
    sys.exit()


def validate_schemas():
    check_schema(data_helper.get_schema_file('auth'))
    check_schema(data_helper.get_schema_file('collection'))
    check_schema(data_helper.get_schema_file('environment'))
    check_schema(data_helper.get_schema_file('folder'))
    check_schema(data_helper.get_schema_file('request'))
    check_schema(data_helper.get_schema_file('settings'))


def check_schema(_file):
    _dict = read_json(_file)
    validator = Draft7Validator(_dict, registry=create_registry())
    validator.check_schema(_dict)


def read_json(filename):
    if not os.path.exists(filename):
        return None
    try:
        with open(filename, 'r') as f:
            return json.load(f)
    except JSONDecodeError as e:
        print()
        print(type(e).__name__ + " " + str(e))
        raise click.ClickException("unable to load file as JSON: " + filename)


def parse_json(json_string):
    try:
        _json = json.loads(json_string)
    except ValueError:  # includes simplejson.decoder.JSONDecodeError
        return None
    return _json


def write_json(filename, data):
    with open(filename, 'w') as f:
        json.dump(data, f, indent=4)


@rc_print_durations
@rc_memoized
def read_settings():
    home = config_helper.get_config_folder()
    return load_and_validate(SETTINGS_FILENAME, _dir=home)


def write_settings(settings):
    home = config_helper.get_config_folder()
    write_json(os.path.join(home, SETTINGS_FILENAME), settings)


@rc_print_durations
@rc_memoized
def find_current_collection():
    settings = load_and_validate(SETTINGS_FILENAME)
    current_name = settings.get('current_collection', None) or None
    collections = settings.get('collections', None) or None
    if current_name is None or collections is None:
        return None

    for c in collections:
        if c['name'] == current_name:
            return c
    return None


@rc_print_durations
@rc_memoized
def read_current_collection():
    current = find_current_collection()
    c = load_and_validate(COLLECTION_FILENAME, _dir=current['location'])
    return c, {
        '_dir': current['location'],
        '_original': c
    }


def write_collection(c):
    location = c['_dir']
    write_json(os.path.join(location, COLLECTION_FILENAME), c['_original'])


@rc_print_durations
@rc_memoized
def read_request_list():
    c, c_wrapper = read_current_collection()
    root = c_wrapper['_dir']
    current_request = c.get('current_request', "")
    _list = []
    number = 0
    for dirpath, dirnames, files in os.walk(root):
        for file in files:
            if dirpath.endswith('examples'):
                continue
            if file.endswith('.request'):
                number = number + 1
                display_num = str(number)

                short_file = file.split('.')[0]
                short_path = dirpath.replace(root, "")
                short_path = short_path.replace("\\", "/")
                short_request = short_path + "/" + file
                current = False

                if short_request == current_request:
                    current = True
                    display_num += '*'

                # read each request to get the METHOD
                r = load_and_validate(file, _dir=dirpath)
                method = r.get('method', '')

                # edit path further for display
                if len(short_path) > 49:
                    short_path = ".../" + short_path[-46]

                _list.append({
                    '_index': number - 1,
                    '_current': current,
                    '_dir': dirpath,
                    '_filename': file,
                    '_original': r,
                    '_display_ref': str(number) + ":" + short_request,
                    '_short_request': short_request,
                    'number': number,
                    'display_num': display_num,
                    'folder': short_path,
                    'method': method,
                    'name': short_file
                })
    return _list


@rc_print_durations
@rc_memoized
def read_environment_list():
    c, c_wrapper = read_current_collection()
    env_folder = os.path.join(c_wrapper['_dir'], 'environments')
    _list = []
    idx = 0
    for dirpath, dirnames, files in os.walk(env_folder):
        for file in files:
            if file.endswith('.json'):
                env = load_and_validate(file, _dir=dirpath)
                name = file.split('.')[-2]
                idx = idx + 1
                number = str(idx)

                current = False
                if name == c['current_environment']:
                    current = True
                    number += '*'
                _list.append({
                    '_index': idx - 1,
                    '_current': current,
                    '_dir': dirpath,
                    '_filename': file,
                    '_original': env,
                    'name': file.split('.')[-2],
                    'number': idx,
                    'display_num': number,
                    'baseUrl': env.get('baseUrl', 'None')
                })
    return _list


@rc_print_durations
@rc_memoized
def read_collection_list():
    settings = read_settings()
    current_c = find_current_collection()
    _list = []
    for idx, c in enumerate(settings.get('collections', [])):
        number = str(idx + 1)
        current = False
        if c['name'] == current_c['name']:
            current = True
            number += '*'

        _list.append({
            '_index': idx,
            '_current': current,
            'number': str(idx+1),
            'display_num': number,
            'name': c['name'],
            'location': c['location']
        })
    return _list


@rc_print_durations
@rc_memoized
def read_environment(name):
    c, c_wrapper = read_current_collection()
    env_folder = os.path.join(c_wrapper['_dir'], 'environments')
    if name == 'current':
        name = c['current_environment']
    if name == 'global':
        name = 'rc-global'
        env_folder = config_helper.get_config_folder()
    env_filename = name+'.json'
    env = load_and_validate(env_filename, _dir=env_folder)
    return env_filename, env


def write_environment(filename, data):
    if filename == SETTINGS_FILENAME:
        print("wtf?")
        sys.exit(-1)
    _dir = guess_dir(filename)
    write_json(os.path.join(_dir, filename),
               data)


def write_request(r, wrapper):
    path = os.path.join(wrapper.get('_dir'), wrapper.get('_filename'))
    write_json(path, r)





import errno
import json
import socket
import sys

import click

from rc3.common import json_helper


def get_json_string(_dict):
    settings = json_helper.read_settings()
    indent = settings.get('indent', 4)
    indent_type = settings.get('indent_type', "space")
    indent_value = indent
    if indent_type == "tab" and indent > 0:
        indent_value = '\t' * indent
    if indent < 0:
        json_string = json.dumps(_dict)
    else:
        json_string = json.dumps(_dict, indent=indent_value)
    return json_string


def extract_json_string(incoming):
    json_string = incoming[incoming.index("{"):]
    return json_string


def print_json(_dict, err=False):
    json_string = get_json_string(_dict)
    if err:
        print(json_string, file=sys.stderr)
    else:
        print(json_string)


def print_json_or_text(response):
    try:
        _dict = response.json()
        print_json(_dict)
    except socket.error as e:
        if e.errno != errno.EPIPE:
            print_text(response)
        # ignore broken pipes


def print_text(response):
    try:
        if len(response.text) == 0:
            print("Status: " + str(response.status_code))
            print("No response body.")
        else:
            print(response.text)
    except socket.error as e:
        if e.errno != errno.EPIPE:
            raise
        # ignore broken pipes


def left_format(_list, field, header_len):
    max_len = max(len(str(obj[field])) for obj in _list)
    max_len = max(max_len, header_len)
    this_format = '{:<' + str(max_len + 3) + '}'
    return this_format


def print_formatted_table(header, fields, _list):
    # Create line_format based on MAX WIDTH of each field we are printing
    line_format = ""
    for idx, field in enumerate(fields):
        header_len = len(header[idx])
        line_format += left_format(_list, field, header_len)

    # print header & then each row in the list
    print(line_format.format(*header))
    for row in _list:
        # first convert into a list of values (instead of a dict/object)
        new_row = []
        for field in fields:
            new_row.append(row[field])
        # then print it
        print(line_format.format(*new_row))


def test_formatted_table():
    header = ['IDENTIFIER:', 'NAME:']
    fields = ['id', 'name']
    _list = [
        {'id': 1, 'name': 'gary'},
        {'id': 2, 'name': 'bob maloney'}
    ]
    print_formatted_table(header, fields, _list)


if __name__ == '__main__':
    test_formatted_table()



import json
import os
import click

from rc3.commands import cmd_list
from rc3.common import json_helper, print_helper


@click.command("request", short_help="Manage REQUESTS in current_collection.")
@click.option('-l', '--list', '_list', is_flag=True, default=False, help="List REQUESTS in collection.")
@click.option('-n', '--new', is_flag=True, default=False, help="Create a new REQUEST.")
@click.option('-i', '--info', is_flag=True, default=False, help="Display json of current REQUEST.")
@click.option('-p', '--pick', is_flag=True, default=False, help="Pick an item as current REQUEST.")
@click.option('-e', '--edit', is_flag=True, default=False, help="Edit a REQUEST with system editor.")
@click.argument('request_name', type=str, required=False)
def cli(_list, new, info, pick, edit, request_name):
    """\b
    Manages REQUEST objects/files in the current collection.

    \b
    REQUEST_NAME is optional.
    REQUEST_NAME will default to the current_request, or prompt you to pick.
    REQUEST_NAME if used should be one of:
    1. The NUM column from 'rc request --list' output
    2. THe NAME column from 'rc request --list' output

    """
    if _list:
        cmd_list.list_requests()
    elif new:
        click.echo("NOT IMPLEMENTED! Please use VSCode to edit your collection.")
    elif edit:
        edit_request(request_name)
    elif info:
        print_info(request_name)
    elif pick:
        pick_request(request_name)
    elif request_name is None:
        # DEFAULT to list() if no REQUEST_NAME
        cmd_list.list_requests()
    else:
        # DEFAULT to pick() if REQUEST_NAME
        pick_request(request_name)


def pick_request(name=None):
    r = lookup_or_prompt(name)
    if r is None:
        return None

    c, c_wrapper = json_helper.read_current_collection()
    c['current_request'] = r.get('_short_request', '')

    json_helper.write_collection(c_wrapper)
    click.echo("REQUEST has been picked: " + r.get('_display_ref'))
    return r


def edit_request(name, wrapper=None):
    if wrapper is None:
        wrapper = lookup_request(name)
    if wrapper is None:
        raise click.ClickException("REQUEST '{}' not found. See 'rc request --list'".format(name))
    r = internal_edit(wrapper)
    if r is not None:
        save_request(r, wrapper)
        wrapper['_original'] = r
    return wrapper


def internal_edit(wrapper):
    json_string = print_helper.get_json_string(wrapper.get('_original'))
    new_string = click.edit(json_string)
    if new_string is not None:
        r = json_helper.parse_json(new_string)
        if r is None:
            raise click.ClickException("new REQUEST must be valid JSON.")
        if json_helper.validate(r, 'request'):
            return r
    else:
        return None


def lookup_or_prompt(name, prompt="Which REQUEST do you want to pick?"):
    _list = json_helper.read_request_list()
    if len(_list) == 0:
        raise click.ClickException("Invalid option.  There are no REQUESTS defined.")

    picked_object = None
    if name is None:
        cmd_list.list_requests()
        current_r = lookup_request(None)
        default_choice = 0 if current_r is None else current_r.get('number', 0)
        number = click.prompt(prompt, default=default_choice)
        picked_object = lookup_request(str(number))
        if picked_object is None:
            raise click.ClickException("Invalid selection")
    else:
        picked_object = lookup_request(name)
        if picked_object is None:
            raise click.ClickException("REQUEST_NAME not found: " + name)
        else:
            click.echo("Found REQUEST_NAME: " + name)
    return picked_object


def print_info(request_name):
    r = lookup_request(request_name)
    print(r.get('_display_ref'))
    print_helper.print_json(r.get('_original', None))


def lookup_request(name):
    _list = json_helper.read_request_list()

    for r in _list:
        if name is None and r.get('_current', False):
            return r
        if name is not None and name == r.get('name', None):
            return r
        if name is not None and name == str(r.get('number', 0)):
            return r
    else:
        return None


def save_request(r, wrapper):
    json_helper.write_request(r, wrapper)
    print("REQUEST saved: ")
    print(wrapper.get('_display_ref'))




import os
import re

import pytest
from click import ClickException

from rc3 import cli
from rc3.common import json_helper, print_helper, env_helper, inherit_helper
from rc3.common.data_helper import FOLDER_FILENAME
from tests.commands import test_request


def auth_remove():
    return {}


def auth_none():
    return {
        "type": "none"
    }


def auth_inherit():
    return {
        "type": "inherit"
    }


def auth_basic():
    return {
        "type": "basic",
        "username": "gary",
        "password": "is-cool"
    }


def auth_bearer():
    return {
        "type": "bearer",
        "bearer_token": "ABCDEFGHIJKLMNOPQRSTUVWXY..."
    }


def set_collection(auth):
    c, wrapper = json_helper.read_current_collection()
    c['auth'] = auth
    json_helper.write_collection(wrapper)


def set_folder(auth, _dir=None):
    if _dir is None:
        r, wrapper = test_request.lookup_current()
        _dir = wrapper['_dir']
    full_filename = os.path.join(_dir, FOLDER_FILENAME)
    folder = {}
    if os.path.exists(full_filename):
        folder = json_helper.load_and_validate(FOLDER_FILENAME, _dir=_dir)

    folder['auth'] = auth
    json_helper.write_json(full_filename, folder)


def delete_folder(_dir=None):
    if _dir is None:
        r, wrapper = test_request.lookup_current()
        _dir = wrapper['_dir']
    full_filename = os.path.join(_dir, FOLDER_FILENAME)
    if os.path.exists(full_filename):
        os.remove(full_filename)


def set_request(auth):
    r, wrapper = test_request.lookup_current()
    r['auth'] = auth
    json_helper.write_request(r, wrapper)


def test_stops_at_request_none(example_collection, runner):
    set_request(auth_none())
    set_folder(auth_basic())
    set_collection(auth_basic())

    r, wrapper = test_request.lookup_current()
    auth = inherit_helper.find_auth(wrapper)

    assert auth.get('type') == "none"


def test_stops_at_folder_basic(example_collection, runner):
    set_request(auth_remove())
    set_folder(auth_basic())
    set_collection(auth_bearer())

    r, wrapper = test_request.lookup_current()
    auth = inherit_helper.find_auth(wrapper)

    assert auth.get('type') == "basic"


def test_follows_explicit_inherit(example_collection, runner):
    set_request(auth_inherit())
    set_folder(auth_inherit())
    set_collection(auth_bearer())

    r, wrapper = test_request.lookup_current()
    auth = inherit_helper.find_auth(wrapper)

    assert auth.get('type') == "bearer"


def test_follows_implicit_inherit(example_collection, runner):
    set_request(auth_remove())
    set_folder(auth_remove())
    set_collection(auth_bearer())

    r, wrapper = test_request.lookup_current()
    auth = inherit_helper.find_auth(wrapper)

    assert auth.get('type') == "bearer"


def test_follows_missing_folder(example_collection, runner):
    set_request(auth_inherit())
    delete_folder()
    set_collection(auth_bearer())

    r, wrapper = test_request.lookup_current()
    auth = inherit_helper.find_auth(wrapper)

    assert auth.get('type') == "bearer"


def test_stops_at_collection(example_collection, runner):
    set_request(auth_inherit())
    set_folder(auth_inherit())
    set_collection(auth_inherit())

    r, wrapper = test_request.lookup_current()
    auth = inherit_helper.find_auth(wrapper)

    assert auth.get('type') == "none"





import collections
import collections.abc
import os

from funcy import print_durations, decorator

collections.Hashable = collections.abc.Hashable


@decorator
def rc_print_durations(call):
    rc_durations = os.getenv("RC_DURATIONS", 'False').lower() in ('true', '1')
    if not rc_durations:
        return call()

    with print_durations(call._func.__name__, unit="ms"):
        return call()


super_cache={}
@decorator
def rc_memoized(call):
    func_name = call._func.__name__
    args = call._args
    no_cache = os.getenv("RC_NO_CACHE", 'False').lower() in ('true', '1')
    if no_cache:
        return call()
    if not isinstance(args, collections.Hashable):
        return call()

    my_cache = super_cache.get(func_name)
    if my_cache is None:
        my_cache = super_cache[func_name] = {}

    if args in my_cache:
        return my_cache[args]
    else:
        my_cache[args] = call()
        return my_cache[args]




import os
import re

import click
import requests
from jsonpath_ng import parse
from requests.auth import HTTPBasicAuth, AuthBase
from requests.exceptions import SSLError

from rc3.commands import cmd_request
from rc3.common import json_helper, print_helper, env_helper, inherit_helper, rc_globals


@click.command("send", short_help="Sends an HTTP request & writes results to a response file.")
@click.option('-p', '--pick', is_flag=True, default=False, help="Pick a REQUEST then send it.")
@click.option('-e', '--edit', is_flag=True, default=False, help="Edit a REQUEST then send it.")
@click.argument('request_name', type=str, required=False)
def cli(pick, edit, request_name):
    """\b
    Will send an HTTP request using the Python requests library.
    \b
    REQUEST_NAME is optional.
    REQUEST_NAME will default to the current_request.
    REQUEST_NAME if used should be one of:
    1. The NUM column from 'rc request --list' output
    2. THe NAME column from 'rc request --list' output

    \b
    OUTPUT will be the response body to STDOUT.
    OUTPUT will also be a *.response file in the same dir as the *.request file.
    """
    if pick and edit:
        r = cmd_request.pick_request(request_name)
        r = cmd_request.edit_request(None, wrapper=r)
        send(r)
    elif pick:
        r = cmd_request.pick_request(request_name)
        send(r)
    elif edit:
        r = cmd_request.edit_request(request_name)
        send(r)
    else:
        r = lookup_request(request_name)
        send(r)


def lookup_request(request_name):
    r = cmd_request.lookup_request(request_name)
    if r is None and request_name is None:
        raise click.ClickException("There is no current REQUEST, exiting...")
    if r is None:
        raise click.ClickException("REQUEST '{}' not found. See 'rc request --list'".format(request_name))
    return r


def send(wrapper):
    settings = json_helper.read_settings()
    request = wrapper.get('_original')
    headers = request.get('headers', {})

    # find_auth() before var substitution!
    request['auth'] = inherit_helper.find_auth(wrapper)
    env_helper.process_subs(wrapper)

    body_count = 0
    _json = request.get('body', {}).get('json', None)
    if _json is not None:
        body_count += 1
    text = request.get('body', {}).get('text', None)
    if text is not None:
        body_count += 1
    form_data = request.get('form_data', None)
    if form_data is not None:
        body_count += 1
    if body_count > 1:
        raise click.ClickException("REQUEST can only have 1 of body.json, body.text, or form_data.  You have {}.".format(body_count))

    _data = None
    if text is not None:
        _data = text
        headers['Content-Type'] = "text/plain"
    elif form_data is not None:
        _data = form_data

    timeout = settings.get('request_timeout', 30)
    allow_redirects = settings.get('follow_redirects', False)
    verify = settings.get('ca_cert_verification', True)
    if verify and len(settings.get('ca_bundle','')) > 0:
        verify = settings.get('ca_bundle')
        if not os.path.exists(verify):
            raise click.ClickException(f'settings.json ca_bundle file [{verify}] does not exist!')
    if settings.get('headers_send_nocache', True):
        headers['Cache-Control'] = 'no-cache'

    try:
        response = requests.request(request.get('method'), request.get('url'),
                                    headers=headers,
                                    json=_json,
                                    auth=create_auth(inherit_helper.find_auth(wrapper)),
                                    params=request.get('params', None),
                                    data=_data,
                                    timeout=timeout,
                                    allow_redirects=allow_redirects,
                                    verify=verify)
    except SSLError as e:
        print()
        print(type(e).__name__ + ": " + str(e))
        click.echo("SSLError: Try setting REQUESTS_CA_BUNDLE, CURL_CA_BUNDLE, or rc-settings.ca_bundle")
        click.echo("See: https://github.com/gswilcox01/rc3/tree/master?tab=readme-ov-file#ca-certificates")
        raise click.Abort

    process_output(wrapper, response)


def process_output(wrapper, response):
    settings = json_helper.read_settings()
    cli_options = rc_globals.get_cli_options()
    request = wrapper.get('_original')

    # process potential extract, and update env file
    extract_errors = process_extract(request, response)

    # determine verbose output for console or file
    verbose_output = create_verbose_output(response)
    verbose_output['extract_errors'] = extract_errors

    # write out *.response file
    save_responses = request.get('save_responses', settings.get('save_responses', True))
    if save_responses:
        response_filename = os.path.join(wrapper.get('_dir'),
                                         wrapper.get('_filename').split('.')[0] + ".response")
        json_helper.write_json(response_filename, verbose_output)

    # display either verbose_output OR just the response body
    if cli_options.get('verbose', False):
        print_helper.print_json(verbose_output)
    else:
        print_helper.print_json_or_text(response)


def process_extract(request, response):
    # collect extract errors, and return them for display or storage on .response
    errors = []

    # do nothing if not "extract" in the def
    extract = request.get("extract", None)
    if extract is None:
        return errors

    # sys.exit with error, if env doesn't exist in the collection (or global doesn't exist)
    env_name = extract.get("env", "global")
    env_filename, env = json_helper.read_environment(env_name)

    # default to "token" if var is not set
    var = extract.get("var", "token")

    # attempt to find json path
    # See: https://www.digitalocean.com/community/tutorials/python-jsonpath-examples
    # See: https://jsonpath.com/
    # See: https://pypi.org/project/jsonpath-ng/
    json_path = extract.get("json_path", None)
    if json_path is not None:
        _json = extract_json(response)
        if _json is None:
            errors.append("REQUEST has $.extract.json_path, but response is not JSON!")
            return errors
        jsonpath_expression = parse(json_path)
        match = jsonpath_expression.find(_json)
        if len(match) < 1:
            errors.append("REQUEST json_path=[{}] had no matches in response JSON!".format(json_path))
            return errors
        # use the value from the first match
        value = match[0].value
        env[var] = value
        json_helper.write_environment(env_filename, env)
        return errors

    # See python regex docs here for valid regex patterns:
    # https://docs.python.org/3/howto/regex.html
    text_pattern = extract.get("text_pattern", None)
    if text_pattern is not None:
        text = response.text
        if len(text) < 1:
            errors.append("REQUEST has $.extract.text_pattern, but response.text is 0 length!")
            return errors
        if "(" not in text_pattern or ")" not in text_pattern:
            errors.append("REQUEST $.extract.text_pattern, MUST contain a matching group ()!")
            return errors
        # do it
        pattern = re.compile(text_pattern)
        match = pattern.search(text)
        # group(0) is the WHOLE match, (1) is the first group in the pattern (what we want)
        if match is None:
            errors.append("REQUEST text_pattern=[{}] had no matches in response JSON!".format(text_pattern))
            return errors
        env[var] = match.group(1)
        json_helper.write_environment(env_filename, env)
        return errors


def extract_json(response):
    try:
        _json = response.json()
        return _json
    except BaseException as error:
        return None


def create_verbose_output(response):
    body = {}
    if extract_json(response) is None:
        body['text'] = response.text
    else:
        body['json'] = extract_json(response)

    headers = {}
    header_size = 0
    for key, value in response.headers.items():
        header_size += len(key) + len(value)
        headers[key] = value

    # verbose output
    return {
        "status_code": response.status_code,
        "time": str(response.elapsed.microseconds / 1000) + "ms",
        "size": {
            "body": len(response.content),
            "headers": header_size,
            "total": len(response.content) + header_size
        },
        "headers": headers,
        "body": body
    }


class HTTPTokenAuth(AuthBase):
    def __init__(self, token, token_header="Authorization", token_name="Bearer"):
        self.token = token
        self.token_header = token_header
        self.token_name = token_name

    def __call__(self, r):
        value = self.token_name + " " + self.token
        r.headers[self.token_header] = value.strip()
        return r


def create_auth(auth_config):
    _auth = None
    _type = auth_config.get('type')
    if _type == 'basic':
        _auth = HTTPBasicAuth(auth_config.get('username'), auth_config.get('password'))
    if _type == 'bearer':
        _auth = HTTPTokenAuth(auth_config.get('bearer_token'))
    if _type == 'token':
        _auth = HTTPTokenAuth(auth_config.get('token_value'),
                              token_name="",
                              token_header=auth_config.get('token_header', "Authorization"))
    return _auth



import os
import click

from rc3.common import json_helper, config_helper, rc_globals
from rc3.common.data_helper import SETTINGS_FILENAME

cmd_folder = os.path.abspath(os.path.join(os.path.dirname(__file__), "commands"))


# based on: https://github.com/pallets/click/blob/main/examples/complex/complex/cli.py
class ComplexCLI(click.Group):
    def list_commands(self, ctx):
        rv = []
        for filename in os.listdir(cmd_folder):
            if filename.endswith(".py") and filename.startswith("cmd_"):
                rv.append(filename[4:-3])
        rv.sort()
        return rv

    def get_command(self, ctx, name):
        # See more options like unique prefix matching here:
        # https://click.palletsprojects.com/en/8.1.x/advanced/#command-aliases
        aliases = {
            'r': 'request',
            'c': 'collection',
            'e': 'environment',
            'g': 'global',
            'globals': 'global'
        }
        name = aliases.get(name, name)
        try:
            mod = __import__(f"rc3.commands.cmd_{name}", None, None, ["cli"])
        except ImportError:
            return
        return mod.cli


@click.command(cls=ComplexCLI)
@click.pass_context
@click.option('-v', '--verbose', is_flag=True, default=False, help="Verbose output.")
def cli(ctx, verbose):
    """A REST CLI for configuring & executing COLLECTIONS of REQUESTS
    """

    # add to global cli_options
    cli_options = rc_globals.get_cli_options()
    cli_options['verbose'] = verbose

    # validate all the schemas in the project
    json_helper.validate_schemas()

    # Note: this next cmd will always create RC_HOME / ~/.rc if it doesn't exist
    home = config_helper.get_config_folder()
    dest = os.path.join(home, SETTINGS_FILENAME)
    if os.path.exists(dest):
        # validate the RC/settings.json file (& sys.exit() if invalid)
        json_helper.load_and_validate(SETTINGS_FILENAME)





import functools
import inspect
import os

import responses
from responses._recorder import Recorder


def default_filename(func):
    module = inspect.getmodule(func)

    directory = os.path.splitext(module.__file__)[0] + "_files"
    os.makedirs(directory, exist_ok=True)

    filename = func.__name__ + ".yaml"
    return os.path.join(directory, filename)


def activate_responses(file_path=None):
    def inner_decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            nonlocal file_path
            if file_path is None:
                file_path = default_filename(func)

            with responses.RequestsMock() as rsp:
                rsp._add_from_file(file_path=file_path)
                func(*args, **kwargs)

        return wrapper

    return inner_decorator


def activate_recorder(file_path=None):
    def inner_decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            nonlocal file_path
            if file_path is None:
                file_path = default_filename(func)

            recorder = Recorder()
            with recorder:
                result = func(*args, **kwargs)
                recorder.dump_to_file(
                    file_path=file_path, registered=recorder.get_registry().registered
                )
                return result

        return wrapper

    return inner_decorator
